# -*- python -*-
# -*- coding: utf-8 -*-
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2011-2016  Serge Noiraud
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Geography for places
"""
# -------------------------------------------------------------------------
#
# Python modules
#
# -------------------------------------------------------------------------
import time
import operator
from collections import defaultdict

# -------------------------------------------------------------------------
#
# set up logging
#
# -------------------------------------------------------------------------
import logging

# -------------------------------------------------------------------------
#
# Gramps Modules
#
# -------------------------------------------------------------------------
from gi.repository import Gdk
from gi.repository import Gtk
from gramps.gen.const import GRAMPS_LOCALE as glocale
from gramps.gen.lib import EventType
from gramps.gen.lib import PlaceType
from gramps.gen.config import config
from gramps.gen.display.place import displayer as _pd
from gramps.gen.utils.place import conv_lat_lon
from gramps.gui.views.bookmarks import PlaceBookmarks
from gramps.plugins.lib.maps.geography import GeoGraphyView
from gramps.plugins.lib.maps import constants
from gramps.gui.utils import ProgressMeter

# -------------------------------------------------------------------------
#
# Constants
#
# -------------------------------------------------------------------------
KEY_TAB = Gdk.KEY_Tab
_ = glocale.translation.gettext
_LOG = logging.getLogger("GeoGraphy.geoplaces")

_UI_DEF = [
    """
      <placeholder id="CommonGo">
      <section>
        <item>
          <attribute name="action">win.Back</attribute>
          <attribute name="label" translatable="yes">_Back</attribute>
        </item>
        <item>
          <attribute name="action">win.Forward</attribute>
          <attribute name="label" translatable="yes">_Forward</attribute>
        </item>
      </section>
      </placeholder>
""",
    """
      <section id='CommonEdit' groups='RW'>
        <item>
          <attribute name="action">win.PrintView</attribute>
          <attribute name="label" translatable="yes">Print...</attribute>
        </item>
      </section>
""",
    """
      <section id="AddEditBook">
        <item>
          <attribute name="action">win.AddBook</attribute>
          <attribute name="label" translatable="yes">_Add Bookmark</attribute>
        </item>
        <item>
          <attribute name="action">win.EditBook</attribute>
          <attribute name="label" translatable="no">%s...</attribute>
        </item>
      </section>
"""
    % _("Organize Bookmarks"),  # Following are the Toolbar items
    """
    <placeholder id='CommonNavigation'>
    <child groups='RO'>
      <object class="GtkToolButton">
        <property name="icon-name">go-previous</property>
        <property name="action-name">win.Back</property>
        <property name="tooltip_text" translatable="yes">"""
    """Go to the previous object in the history</property>
        <property name="label" translatable="yes">_Back</property>
        <property name="use-underline">True</property>
      </object>
      <packing>
        <property name="homogeneous">False</property>
      </packing>
    </child>
    <child groups='RO'>
      <object class="GtkToolButton">
        <property name="icon-name">go-next</property>
        <property name="action-name">win.Forward</property>
        <property name="tooltip_text" translatable="yes">"""
    """Go to the next object in the history</property>
        <property name="label" translatable="yes">_Forward</property>
        <property name="use-underline">True</property>
      </object>
      <packing>
        <property name="homogeneous">False</property>
      </packing>
    </child>
    </placeholder>
""",
    """
    <placeholder id='BarCommonEdit'>
    <child groups='RO'>
      <object class="GtkToolButton">
        <property name="icon-name">document-print</property>
        <property name="action-name">win.PrintView</property>
        <property name="tooltip_text" translatable="yes">"""
    """Print or save the Map</property>
        <property name="label" translatable="yes">Print...</property>
        <property name="use-underline">True</property>
      </object>
      <packing>
        <property name="homogeneous">False</property>
      </packing>
    </child>
    </placeholder>
    """,
]

# pylint: disable=no-member
# pylint: disable=maybe-no-member
# pylint: disable=unused-variable
# pylint: disable=unused-argument


# -------------------------------------------------------------------------
#
# GeoView
#
# -------------------------------------------------------------------------
class GeoPlaces(GeoGraphyView):
    """
    The view used to render places map.
    """

    CONFIGSETTINGS = (
        ("geography.path", constants.GEOGRAPHY_PATH),
        ("geography.zoom", 10),
        ("geography.zoom_when_center", 12),
        ("geography.show_cross", True),
        ("geography.lock", False),
        ("geography.center-lat", 0.0),
        ("geography.center-lon", 0.0),
        ("geography.map_service", constants.OPENSTREETMAP),
        ("geography.max_places", 5000),
        ("geography.use-keypad", True),
        ("geography.personal-map", ""),
        # specific to geoplaces :
        ("geography.color.unknown", "#008b00"),
        ("geography.color.custom", "#008b00"),
        ("geography.color.country", "#008b00"),
        ("geography.color.county", "#008b00"),
        ("geography.color.state", "#008b00"),
        ("geography.color.city", "#008b00"),
        ("geography.color.parish", "#008b00"),
        ("geography.color.locality", "#008b00"),
        ("geography.color.street", "#008b00"),
        ("geography.color.province", "#008b00"),
        ("geography.color.region", "#008b00"),
        ("geography.color.department", "#008b00"),
        ("geography.color.neighborhood", "#008b00"),
        ("geography.color.district", "#008b00"),
        ("geography.color.borough", "#008b00"),
        ("geography.color.municipality", "#008b00"),
        ("geography.color.town", "#008b00"),
        ("geography.color.village", "#008b00"),
        ("geography.color.hamlet", "#008b00"),
        ("geography.color.farm", "#008b00"),
        ("geography.color.building", "#008b00"),
        ("geography.color.number", "#008b00"),
    )

    def __init__(self, pdata, dbstate, uistate, nav_group=0):
        self.window_name = _("Places map")
        GeoGraphyView.__init__(
            self, self.window_name, pdata, dbstate, uistate, PlaceBookmarks, nav_group
        )
        self.dbstate = dbstate
        self.uistate = uistate
        self.place_list = []
        self.place_without_coordinates = []
        self.minlat = self.maxlat = self.minlon = self.maxlon = 0.0
        self.minyear = 9999
        self.maxyear = 0
        self.nbplaces = 0
        self.nbmarkers = 0
        self.sort = []
        self.generic_filter = None
        self.additional_uis.append(self.additional_ui())
        self.no_show_places_in_status_bar = False
        self.show_all = False
        self.itemoption = None
        self.menu = None
        self.cal = config.get("preferences.calendar-format-report")
        self.plc_color = []
        self.plc_custom_color = defaultdict(set)

    def get_title(self):
        """
        Used to set the titlebar in the configuration window.
        """
        return _("GeoPlaces")

    def get_stock(self):
        """
        Returns the name of the stock icon to use for the display.
        This assumes that this icon has already been registered
        as a stock icon.
        """
        return "geo-show-place"

    def get_viewtype_stock(self):
        """Type of view in category"""
        return "geo-show-place"

    def additional_ui(self):
        """
        Specifies the UIManager XML code that defines the menus and buttons
        associated with the interface.
        """
        return _UI_DEF

    def navigation_type(self):
        """
        Indicates the navigation type. Navigation type can be the string
        name of any of the primary objects.
        """
        return "Place"

    def goto_handle(self, handle=None):
        """
        Rebuild the tree with the given places handle as the root.
        """
        if self.osm is None:
            return
        self.places_found = []
        active = self.uistate.get_active("Place")
        if active:
            self._createmap(active)
        else:
            self._createmap(None)

    def show_all_places(self, menu, event, lat, lon):
        """
        Ask to show all places.
        """
        self.show_all = True
        self.nbmarkers = 0
        self._createmap(None)

    def build_tree(self):
        """
        This is called by the parent class when the view becomes visible. Since
        all handling of visibility is now in rebuild_trees, see that for more
        information.
        """
        pass

    def _create_one_place(self, place):
        """
        Create one entry for one place with a lat/lon.
        """
        if place is None:
            return
        if self.nbplaces >= self._config.get("geography.max_places"):
            return
        descr = _pd.display(self.dbstate.db, place)
        longitude = place.get_longitude()
        latitude = place.get_latitude()
        latitude, longitude = conv_lat_lon(latitude, longitude, "D.D8")
        self.load_kml_files(place)
        # place.get_longitude and place.get_latitude return
        # one string. We have coordinates when the two values
        # contains non null string.
        if longitude and latitude:
            colour = self.plc_color[int(place.get_type()) + 1]
            if int(place.get_type()) == PlaceType.CUSTOM:
                try:
                    colour = (
                        str(place.get_type()),
                        self.plc_custom_color[str(place.get_type())],
                    )
                except Exception:
                    colour = self.plc_color[PlaceType.CUSTOM + 1]
            self._append_to_places_list(
                descr,
                None,
                "",
                latitude,
                longitude,
                None,
                None,
                EventType.UNKNOWN,
                None,  # person.gramps_id
                place.gramps_id,
                None,  # event.gramps_id
                None,  # family.gramps_id
                color=colour,
            )

    def _createmap(self, place_x):
        """
        Create all markers for each people's event in the database which has
        a lat/lon.
        """
        if self.osm is None:
            return
        dbstate = self.dbstate
        self.place_list = []
        self.places_found = []
        self.place_without_coordinates = []
        self.minlat = 0.0
        self.maxlat = 0.0
        self.minlon = 0.0
        self.maxlon = 0.0
        self.minyear = 9999
        self.maxyear = 0
        self.without = 0
        latitude = ""
        longitude = ""
        self.nbmarkers = 0
        self.nbplaces = 0
        self.remove_all_markers()
        self.message_layer.clear_messages()
        self.message_layer.clear_font_attributes()
        self.kml_layer.clear()
        self.no_show_places_in_status_bar = False
        _col = self._config.get
        self.plc_color = [
            (PlaceType.UNKNOWN, _col("geography.color.unknown")),
            (PlaceType.CUSTOM, _col("geography.color.custom")),
            (PlaceType.COUNTRY, _col("geography.color.country")),
            (PlaceType.STATE, _col("geography.color.state")),
            (PlaceType.COUNTY, _col("geography.color.county")),
            (PlaceType.CITY, _col("geography.color.city")),
            (PlaceType.PARISH, _col("geography.color.parish")),
            (PlaceType.LOCALITY, _col("geography.color.locality")),
            (PlaceType.STREET, _col("geography.color.street")),
            (PlaceType.PROVINCE, _col("geography.color.province")),
            (PlaceType.REGION, _col("geography.color.region")),
            (PlaceType.DEPARTMENT, _col("geography.color.department")),
            (PlaceType.NEIGHBORHOOD, _col("geography.color.neighborhood")),
            (PlaceType.DISTRICT, _col("geography.color.district")),
            (PlaceType.BOROUGH, _col("geography.color.borough")),
            (PlaceType.MUNICIPALITY, _col("geography.color.municipality")),
            (PlaceType.TOWN, _col("geography.color.town")),
            (PlaceType.VILLAGE, _col("geography.color.village")),
            (PlaceType.HAMLET, _col("geography.color.hamlet")),
            (PlaceType.FARM, _col("geography.color.farm")),
            (PlaceType.BUILDING, _col("geography.color.building")),
            (PlaceType.NUMBER, _col("geography.color.number")),
        ]
        # base "villes de france" : 38101 places :
        # createmap : 8'50"; create_markers : 1'23"
        # base "villes de france" : 38101 places :
        # createmap : 8'50"; create_markers : 0'07" with pixbuf optimization
        # base "villes de france" : 38101 places :
        # gramps 3.4 python 2.7 (draw_markers is estimated when moving the map)
        # 38101 places: createmap: 04'32";
        #               create_markers: 0'04"; draw markers: N/A :: 0'03"
        # 65598 places: createmap: 10'03";
        #               create_markers: 0'07"; draw markers: N/A :: 0'05"
        # gramps 3.5 python 2.7 new marker layer
        # 38101 places: createmap: 03'09";
        #               create_markers: 0'01"; draw markers: 0'04"
        # 65598 places: createmap: 08'48";
        #               create_markers: 0'01"; draw markers: 0'07"
        _LOG.debug(
            "%s",
            time.strftime("start createmap : " "%a %d %b %Y %H:%M:%S", time.gmtime()),
        )
        self.custom_places()
        if self.show_all:
            self.show_all = False
            try:
                places_handle = dbstate.db.get_place_handles()
            except Exception:
                return
            progress = ProgressMeter(
                self.window_name, can_cancel=False, parent=self.uistate.window
            )
            length = len(places_handle)
            progress.set_pass(_("Selecting all places"), length)
            for place_hdl in places_handle:
                place = dbstate.db.get_place_from_handle(place_hdl)
                self._create_one_place(place)
                progress.step()
            progress.close()
        elif self.generic_filter:
            user = self.uistate.viewmanager.user
            place_list = self.generic_filter.apply(dbstate.db, user=user)
            progress = ProgressMeter(
                self.window_name, can_cancel=False, parent=self.uistate.window
            )
            length = len(place_list)
            progress.set_pass(_("Selecting all places"), length)
            for place_handle in place_list:
                place = dbstate.db.get_place_from_handle(place_handle)
                self._create_one_place(place)
                progress.step()
            progress.close()
            # reset completely the filter. It will be recreated next time.
            self.generic_filter = None
        elif place_x is not None:
            place = dbstate.db.get_place_from_handle(place_x)
            self._create_one_place(place)
            self.message_layer.add_message(
                _(
                    "Right click on the map and select 'show all places'"
                    " to show all known places with coordinates. "
                    "You can change the markers color depending on place type. "
                    "You can use filtering."
                )
            )
            if place.get_latitude() != "" and place.get_longitude() != "":
                latitude, longitude = conv_lat_lon(
                    place.get_latitude(), place.get_longitude(), "D.D8"
                )
                if latitude and longitude:
                    zwc = int(config.get("geography.zoom_when_center"))
                    self.osm.set_center_and_zoom(float(latitude), float(longitude), zwc)
        else:
            self.message_layer.add_message(
                _(
                    "Right click on the map and select 'show all places'"
                    " to show all known places with coordinates. "
                    "You can use the history to navigate on the map. "
                    "You can change the markers color depending on place type. "
                    "You can use filtering."
                )
            )
        _LOG.debug(" stop createmap.")
        _LOG.debug(
            "%s", time.strftime("begin sort : " "%a %d %b %Y %H:%M:%S", time.gmtime())
        )
        self.sort = sorted(self.place_list, key=operator.itemgetter(0))
        _LOG.debug(
            "%s", time.strftime("  end sort : " "%a %d %b %Y %H:%M:%S", time.gmtime())
        )
        if self.nbmarkers > 500:  # performance issue. Is it the good value ?
            self.message_layer.add_message(
                _("The place name in the status bar is disabled.")
            )
            self.no_show_places_in_status_bar = True
        if self.nbplaces >= self._config.get("geography.max_places"):
            self.message_layer.set_font_attributes(None, None, "red")
            self.message_layer.add_message(
                _("The maximum number of places is reached (%d).")
                % self._config.get("geography.max_places")
            )
            self.message_layer.add_message(_("Some information are missing."))
            self.message_layer.add_message(
                _("Please, use filtering to reduce this number.")
            )
            self.message_layer.add_message(
                _("You can modify this value in the geography option.")
            )
            self.message_layer.add_message(
                _("In this case, it may take time to show all markers.")
            )

        self._create_markers()

    def bubble_message(self, event, lat, lon, marks):
        self.menu = Gtk.Menu()
        menu = self.menu
        message = ""
        prevmark = None
        for mark in marks:
            if message != "":
                add_item = Gtk.MenuItem(label=message)
                add_item.show()
                menu.append(add_item)
                self.itemoption = Gtk.Menu()
                itemoption = self.itemoption
                itemoption.show()
                add_item.set_submenu(itemoption)
                modify = Gtk.MenuItem(label=_("Edit Place"))
                modify.show()
                modify.connect("activate", self.edit_place, event, lat, lon, prevmark)
                itemoption.append(modify)
                center = Gtk.MenuItem(label=_("Center on this place"))
                center.show()
                center.connect("activate", self.center_here, event, lat, lon, prevmark)
                itemoption.append(center)
                place = self.dbstate.db.get_place_from_gramps_id(mark[9])
                hdle = place.get_handle()
                bookm = Gtk.MenuItem(label=_("Bookmark this place"))
                bookm.show()
                bookm.connect("activate", self.add_bookmark_from_popup, hdle)
                itemoption.append(bookm)
            message = "%s" % mark[0]
            prevmark = mark
        add_item = Gtk.MenuItem(label=message)
        add_item.show()
        menu.append(add_item)
        self.itemoption = Gtk.Menu()
        itemoption = self.itemoption
        itemoption.show()
        add_item.set_submenu(itemoption)
        modify = Gtk.MenuItem(label=_("Edit Place"))
        modify.show()
        modify.connect("activate", self.edit_place, event, lat, lon, prevmark)
        itemoption.append(modify)
        center = Gtk.MenuItem(label=_("Center on this place"))
        center.show()
        center.connect("activate", self.center_here, event, lat, lon, prevmark)
        itemoption.append(center)
        place = self.dbstate.db.get_place_from_gramps_id(mark[9])
        hdle = place.get_handle()
        bookm = Gtk.MenuItem(label=_("Bookmark this place"))
        bookm.show()
        bookm.connect("activate", self.add_bookmark_from_popup, hdle)
        itemoption.append(bookm)
        menu.popup_at_pointer(event)
        return 1

    def add_specific_menu(self, menu, event, lat, lon):
        """
        Add specific entry to the navigation menu.
        """
        menu.append(Gtk.SeparatorMenuItem())

        add_item = Gtk.MenuItem(label=_("Show all places"))
        add_item.connect("activate", self.show_all_places, event, lat, lon)
        menu.append(add_item)
        add_item = Gtk.MenuItem(label=_("Centering on Place"))
        menu.append(add_item)
        self.itemoption = Gtk.Menu()
        itemoption = self.itemoption
        add_item.set_submenu(itemoption)
        oldplace = ""
        for mark in self.sort:
            if mark[0] != oldplace:
                oldplace = mark[0]
                modify = Gtk.MenuItem(label=mark[0])
                modify.connect(
                    "activate", self.goto_place, float(mark[3]), float(mark[4])
                )
                itemoption.append(modify)

    def goto_place(self, obj, lat, lon):
        """
        Center the map on latitude, longitude.
        """
        self.set_center(None, None, lat, lon)

    def get_default_gramplets(self):
        """
        Define the default gramplets for the sidebar and bottombar.
        """
        return (("Place Filter",), ())

    def specific_options(self, configdialog):
        """
        Add specific entry to the preference menu.
        Must be done in the associated view.
        """
        grid = Gtk.Grid()
        grid.set_border_width(12)
        grid.set_column_spacing(6)
        grid.set_row_spacing(6)
        configdialog.add_color(grid, _("Unknown"), 1, "geography.color.unknown", col=1)
        configdialog.add_color(grid, _("Custom"), 2, "geography.color.custom", col=1)
        configdialog.add_color(
            grid, _("Locality"), 3, "geography.color.locality", col=1
        )
        configdialog.add_color(grid, _("Street"), 4, "geography.color.street", col=1)
        configdialog.add_color(
            grid, _("Neighborhood"), 5, "geography.color.neighborhood", col=1
        )
        configdialog.add_color(grid, _("Borough"), 6, "geography.color.borough", col=1)
        configdialog.add_color(grid, _("Village"), 7, "geography.color.village", col=1)
        configdialog.add_color(grid, _("Hamlet"), 8, "geography.color.hamlet", col=1)
        configdialog.add_color(grid, _("Farm"), 9, "geography.color.farm", col=1)
        configdialog.add_color(
            grid, _("Building"), 10, "geography.color.building", col=1
        )
        configdialog.add_color(grid, _("Number"), 11, "geography.color.number", col=1)
        configdialog.add_color(grid, _("Country"), 1, "geography.color.country", col=4)
        configdialog.add_color(grid, _("State"), 2, "geography.color.state", col=4)
        configdialog.add_color(grid, _("County"), 3, "geography.color.county", col=4)
        configdialog.add_color(
            grid, _("Province"), 4, "geography.color.province", col=4
        )
        configdialog.add_color(grid, _("Region"), 5, "geography.color.region", col=4)
        configdialog.add_color(
            grid, _("Department"), 6, "geography.color.department", col=4
        )
        configdialog.add_color(
            grid, _("District"), 7, "geography.color.district", col=4
        )
        configdialog.add_color(grid, _("Parish"), 8, "geography.color.parish", col=4)
        configdialog.add_color(grid, _("City"), 9, "geography.color.city", col=4)
        configdialog.add_color(grid, _("Town"), 10, "geography.color.town", col=4)
        configdialog.add_color(
            grid, _("Municipality"), 11, "geography.color.municipality", col=4
        )
        self.custom_places()
        if len(self.plc_custom_color) > 0:
            configdialog.add_text(grid, _("Custom places name"), 12)
            start = 13
            for color in self.plc_custom_color.keys():
                cust_col = "geography.color." + color.lower()
                row = start if start % 2 else start - 1
                column = 1 if start % 2 else 4
                configdialog.add_color(grid, color, row, cust_col, col=column)
                start += 1
        return _("The places marker color"), grid

    def custom_places(self):
        """
        looking for custom places
        if not registered, register it.
        """
        self.plc_custom_color = defaultdict(set)
        for place in self.dbstate.db.iter_places():
            if int(place.get_type()) == PlaceType.CUSTOM:
                cust_col = "geography.color." + str(place.get_type()).lower()
                try:
                    color = self._config.get(cust_col)
                except Exception:
                    color = "#008b00"
                    self._config.register(cust_col, color)
                if str(place.get_type()) not in self.plc_custom_color.keys():
                    self.plc_custom_color[str(place.get_type())] = color.lower()



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2007-2008  Brian G. Matherly
# Copyright (C) 2008       Gary Burton
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Option class representing a list of filters.
"""

# -------------------------------------------------------------------------
#
# gramps modules
#
# -------------------------------------------------------------------------
from . import EnumeratedListOption


# -------------------------------------------------------------------------
#
# FilterOption class
#
# -------------------------------------------------------------------------
class FilterOption(EnumeratedListOption):
    """
    This class describes an option that provides a list of person filters.
    Each possible value represents one of the possible filters.
    """

    def __init__(self, label, value):
        """
        :param label: A friendly label to be applied to this option.
            Example: "Filter"
        :type label: string
        :param value: A default value for the option.
            Example: 1
        :type label: int
        :return: nothing
        """
        EnumeratedListOption.__init__(self, label, value)
        self.__filters = []

    def set_filters(self, filter_list):
        """
        Set the list of filters available to be chosen from.

        :param filter_list: An array of person filters.
        :type filter_list: array
        :return: nothing
        """
        curval = self.get_value()
        items = [(value, filt.get_name()) for value, filt in enumerate(filter_list)]

        self.__filters = filter_list
        self.clear()
        self.set_items(items)

        self.set_value(curval)

    def get_filter(self):
        """
        Return the currently selected filter object.

        :return: A filter object.
        """
        return self.__filters[self.get_value()]



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Provide the different place types.
"""

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from .grampstype import GrampsType
from ..const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext


class PlaceType(GrampsType):
    UNKNOWN = -1
    CUSTOM = 0
    COUNTRY = 1
    STATE = 2
    COUNTY = 3
    CITY = 4
    PARISH = 5
    LOCALITY = 6
    STREET = 7
    PROVINCE = 8
    REGION = 9
    DEPARTMENT = 10
    NEIGHBORHOOD = 11
    DISTRICT = 12
    BOROUGH = 13
    MUNICIPALITY = 14
    TOWN = 15
    VILLAGE = 16
    HAMLET = 17
    FARM = 18
    BUILDING = 19
    NUMBER = 20

    _CUSTOM = CUSTOM
    _DEFAULT = UNKNOWN

    _DATAMAP = [
        (UNKNOWN, _("Unknown"), "Unknown"),
        (CUSTOM, _("Custom"), "Custom"),
        (COUNTRY, _("Country"), "Country"),
        (STATE, _("State"), "State"),
        (COUNTY, _("County"), "County"),
        (CITY, _("City"), "City"),
        (PARISH, _("Parish"), "Parish"),
        (LOCALITY, _("Locality"), "Locality"),
        (STREET, _("Street"), "Street"),
        (PROVINCE, _("Province"), "Province"),
        (REGION, _("Region"), "Region"),
        (DEPARTMENT, _("Department"), "Department"),
        (NEIGHBORHOOD, _("Neighborhood"), "Neighborhood"),
        (DISTRICT, _("District"), "District"),
        (BOROUGH, _("Borough"), "Borough"),
        (MUNICIPALITY, _("Municipality"), "Municipality"),
        (TOWN, _("Town"), "Town"),
        (VILLAGE, _("Village"), "Village"),
        (HAMLET, _("Hamlet"), "Hamlet"),
        (FARM, _("Farm"), "Farm"),
        (BUILDING, _("Building"), "Building"),
        (NUMBER, _("Number"), "Number"),
    ]

    def __init__(self, value=None):
        GrampsType.__init__(self, value)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2009       Gary Burton
# Copyright (C) 2011       Tim G L Lyons
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Keyword translation interface
"""

# keyword, code, translated standard, translated upper
# in gen.display.name.py we find:
#        't' : title      = title
#        'f' : given      = given (first names)
#        'l' : surname    = full surname (lastname)
#        'c' : call       = callname
#        'x' : common     = nick name if existing, otherwise first first name (common name)
#        'i' : initials   = initials of the first names
#        'm' : primary    = primary surname (main)
#        '0m': primary[pre]= prefix primary surname (main)
#        '1m': primary[sur]= surname primary surname (main)
#        '2m': primary[con]= connector primary surname (main)
#        'y' : patronymic = pa/matronymic surname (father/mother) - assumed unique
#        '0y': patronymic[pre] = prefix      "
#        '1y': patronymic[sur] = surname     "
#        '2y': patronymic[con] = connector   "
#        'o' : notpatronymic = surnames without pa/matronymic and primary
#        'r' : rest       = non primary surnames
#        'p' : prefix     = list of all prefixes
#        'q' : rawsurnames = surnames without prefixes and connectors
#        's' : suffix     = suffix
#        'n' : nickname   = nick name
#        'g' : familynick = family nick name

from ..const import GRAMPS_LOCALE as glocale

_ = glocale.translation.sgettext

KEYWORDS = [
    ("title", "t", _("Title", "Person"), _("TITLE", "Person")),
    ("given", "f", _("Given"), _("GIVEN")),
    ("surname", "l", _("Surname"), _("SURNAME")),
    ("call", "c", _("Call", "Name"), _("CALL", "Name")),
    ("common", "x", _("Common", "Name"), _("COMMON", "Name")),
    ("initials", "i", _("Initials"), _("INITIALS")),
    ("suffix", "s", _("Suffix"), _("SUFFIX")),
    ("primary", "m", _("Primary", "Name"), _("PRIMARY")),
    ("primary[pre]", "0m", _("Primary[pre]"), _("PRIMARY[PRE]")),
    ("primary[sur]", "1m", _("Primary[sur]"), _("PRIMARY[SUR]")),
    ("primary[con]", "2m", _("Primary[con]"), _("PRIMARY[CON]")),
    ("patronymic", "y", _("Patronymic"), _("PATRONYMIC")),
    ("patronymic[pre]", "0y", _("Patronymic[pre]"), _("PATRONYMIC[PRE]")),
    ("patronymic[sur]", "1y", _("Patronymic[sur]"), _("PATRONYMIC[SUR]")),
    ("patronymic[con]", "2y", _("Patronymic[con]"), _("PATRONYMIC[CON]")),
    ("rawsurnames", "q", _("Rawsurnames"), _("RAWSURNAMES")),
    ("notpatronymic", "o", _("Notpatronymic"), _("NOTPATRONYMIC")),
    ("prefix", "p", _("Prefix"), _("PREFIX")),
    ("nickname", "n", _("Nickname"), _("NICKNAME")),
    ("familynick", "g", _("Familynick"), _("FAMILYNICK")),
]
KEY_TO_TRANS = {}
TRANS_TO_KEY = {}
for key, code, standard, upper in KEYWORDS:
    KEY_TO_TRANS[key] = standard
    KEY_TO_TRANS[key.upper()] = upper
    KEY_TO_TRANS["%" + ("%s" % code)] = standard
    KEY_TO_TRANS["%" + ("%s" % code.upper())] = upper
    TRANS_TO_KEY[standard.lower()] = key
    TRANS_TO_KEY[standard] = key
    TRANS_TO_KEY[upper] = key.upper()


def get_translation_from_keyword(keyword):
    """Return the translation of keyword"""
    return KEY_TO_TRANS.get(keyword, keyword)


def get_keyword_from_translation(word):
    """Return the keyword of translation"""
    return TRANS_TO_KEY.get(word, word)


def get_keywords():
    """Get all keywords, longest to shortest"""
    keys = list(KEY_TO_TRANS.keys())
    keys.sort(key=lambda a: len(a), reverse=True)
    return keys


def get_translations():
    """Get all translations, longest to shortest"""
    trans = list(TRANS_TO_KEY.keys())
    trans.sort(key=lambda a: len(a), reverse=True)
    return trans



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2009       Benny Malengier
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
#
"""
Base class for map services. Example: open place in openstreetmap.org
"""

# ------------------------------------------------------------------------
#
# python modules
#
# ------------------------------------------------------------------------

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.utils.place import conv_lat_lon
from gramps.gui.display import display_url


class MapService:
    """Generic base class for map services
    A service is a singleton, we only need one to exist
    Usage is as a callable when used
    """

    def __init__(self):
        self.database = None
        self.items = None
        self.url = ""

    def __call__(self, database, items, uistate):
        """Callable class, usable as a function. This guarantees the class is
        instantiated once when a service is registered. Afterward only calls
        occur
        database: Database to work on
        items: list of tuples (place_handle, description), where description
               is None or a string to use for marker (eg 'birth John Doe')
        uistate: This is only used for transient windows if needed
        """
        self.database = database
        self.items = items
        self.uistate = uistate
        self.url = ""
        # An instance is called, we display the result
        self.calc_url()
        self.__display()
        self._free()

    def _get_first_place(self):
        """Obtain the first place object"""
        place_handle = self.items[0][0]

        return self.database.get_place_from_handle(place_handle), self.items[0][1]

    def _all_places(self):
        """Obtain a list generator of all place objects
        Usage as an iterator:
            for place, descr in mapservice.all_places()
        """
        for handle, descr in self.items:
            yield self.database.get_place_from_handle(handle), descr

    def _lat_lon(self, place, format="D.D8"):
        """return the lat, lon value of place in the requested format
        None, None if invalid
        """
        return conv_lat_lon(place.get_latitude(), place.get_longitude(), format)

    def calc_url(self):
        """Base class needs to overwrite this, calculation of the self.path"""
        raise NotImplementedError

    def __display(self):
        """Show the url in an external browser"""
        if self.url:
            display_url(self.url)

    def _free(self):
        """Allow garbage collection to do it's work"""
        self.items = None



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2010       Michiel D. Nauta
# Copyright (C) 2010,2017  Nick Hall
# Copyright (C) 2011       Tim G L Lyons
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Person object for Gramps.
"""

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from .primaryobj import PrimaryObject
from .citationbase import CitationBase
from .notebase import NoteBase
from .mediabase import MediaBase
from .attrbase import AttributeBase
from .addressbase import AddressBase
from .ldsordbase import LdsOrdBase
from .urlbase import UrlBase
from .tagbase import TagBase
from .name import Name
from .eventref import EventRef
from .personref import PersonRef
from .attrtype import AttributeType
from .eventroletype import EventRoleType
from .attribute import Attribute
from .const import IDENTICAL, EQUAL, DIFFERENT
from ..const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext


# -------------------------------------------------------------------------
#
# Person class
#
# -------------------------------------------------------------------------
class Person(
    CitationBase,
    NoteBase,
    AttributeBase,
    MediaBase,
    AddressBase,
    UrlBase,
    LdsOrdBase,
    PrimaryObject,
):
    """
    The Person record is the Gramps in-memory representation of an
    individual person. It contains all the information related to
    an individual.

    Person objects are usually created in one of two ways.

    1. Creating a new person object, which is then initialized and added to
       the database.
    2. Retrieving an object from the database using the records handle.

    Once a Person object has been modified, it must be committed
    to the database using the database object's commit_person function,
    or the changes will be lost.

    """

    OTHER = 3
    UNKNOWN = 2
    MALE = 1
    FEMALE = 0

    def __init__(self, data=None):
        """
        Create a new Person instance.

        After initialization, most data items have empty or null values,
        including the database handle.
        """
        PrimaryObject.__init__(self)
        CitationBase.__init__(self)
        NoteBase.__init__(self)
        MediaBase.__init__(self)
        AttributeBase.__init__(self)
        AddressBase.__init__(self)
        UrlBase.__init__(self)
        LdsOrdBase.__init__(self)
        self.primary_name = Name()
        self.event_ref_list = []
        self.family_list = []
        self.parent_family_list = []
        self.alternate_names = []
        self.person_ref_list = []
        self.__gender = Person.UNKNOWN
        self.death_ref_index = -1
        self.birth_ref_index = -1
        if data:
            self.unserialize(data)

        # We hold a reference to the GrampsDB so that we can maintain
        # its genderStats.  It doesn't get set here, but from
        # GenderStats.count_person.

    def __eq__(self, other):
        return isinstance(other, Person) and self.handle == other.handle

    def __ne__(self, other):
        return not self == other

    def serialize(self):
        """
        Convert the data held in the Person to a Python tuple that
        represents all the data elements.

        This method is used to convert the object into a form that can easily
        be saved to a database.

        These elements may be primitive Python types (string, integers),
        complex Python types (lists or tuples, or Python objects. If the
        target database cannot handle complex types (such as objects or
        lists), the database is responsible for converting the data into
        a form that it can use.

        :returns: Returns a python tuple containing the data that should
                  be considered persistent.
        :rtype: tuple
        """
        return (
            self.handle,  #  0
            self.gramps_id,  #  1
            self.__gender,  #  2
            self.primary_name.serialize(),  #  3
            [name.serialize() for name in self.alternate_names],  #  4
            self.death_ref_index,  #  5
            self.birth_ref_index,  #  6
            [er.serialize() for er in self.event_ref_list],  #  7
            self.family_list,  #  8
            self.parent_family_list,  #  9
            MediaBase.serialize(self),  # 10
            AddressBase.serialize(self),  # 11
            AttributeBase.serialize(self),  # 12
            UrlBase.serialize(self),  # 13
            LdsOrdBase.serialize(self),  # 14
            CitationBase.serialize(self),  # 15
            NoteBase.serialize(self),  # 16
            self.change,  # 17
            TagBase.serialize(self),  # 18
            self.private,  # 19
            [pr.serialize() for pr in self.person_ref_list],  # 20
        )

    @classmethod
    def get_schema(cls):
        """
        Returns the JSON Schema for this class.

        :returns: Returns a dict containing the schema.
        :rtype: dict
        """
        from .mediaref import MediaRef
        from .address import Address
        from .url import Url
        from .ldsord import LdsOrd

        return {
            "type": "object",
            "title": _("Person"),
            "properties": {
                "_class": {"enum": [cls.__name__]},
                "handle": {"type": "string", "maxLength": 50, "title": _("Handle")},
                "gramps_id": {"type": "string", "title": _("Gramps ID")},
                "gender": {
                    "type": "integer",
                    "minimum": 0,
                    "maximum": 2,
                    "title": _("Gender"),
                },
                "primary_name": Name.get_schema(),
                "alternate_names": {
                    "type": "array",
                    "items": Name.get_schema(),
                    "title": _("Alternate names"),
                },
                "death_ref_index": {
                    "type": "integer",
                    "title": _("Death reference index"),
                },
                "birth_ref_index": {
                    "type": "integer",
                    "title": _("Birth reference index"),
                },
                "event_ref_list": {
                    "type": "array",
                    "items": EventRef.get_schema(),
                    "title": _("Event references"),
                },
                "family_list": {
                    "type": "array",
                    "items": {"type": "string", "maxLength": 50},
                    "title": _("Families"),
                },
                "parent_family_list": {
                    "type": "array",
                    "items": {"type": "string", "maxLength": 50},
                    "title": _("Parent families"),
                },
                "media_list": {
                    "type": "array",
                    "items": MediaRef.get_schema(),
                    "title": _("Media"),
                },
                "address_list": {
                    "type": "array",
                    "items": Address.get_schema(),
                    "title": _("Addresses"),
                },
                "attribute_list": {
                    "type": "array",
                    "items": Attribute.get_schema(),
                    "title": _("Attributes"),
                },
                "urls": {
                    "type": "array",
                    "items": Url.get_schema(),
                    "title": _("Urls"),
                },
                "lds_ord_list": {
                    "type": "array",
                    "items": LdsOrd.get_schema(),
                    "title": _("LDS ordinances"),
                },
                "citation_list": {
                    "type": "array",
                    "items": {"type": "string", "maxLength": 50},
                    "title": _("Citations"),
                },
                "note_list": {
                    "type": "array",
                    "items": {"type": "string", "maxLength": 50},
                    "title": _("Notes"),
                },
                "change": {"type": "integer", "title": _("Last changed")},
                "tag_list": {
                    "type": "array",
                    "items": {"type": "string", "maxLength": 50},
                    "title": _("Tags"),
                },
                "private": {"type": "boolean", "title": _("Private")},
                "person_ref_list": {
                    "type": "array",
                    "items": PersonRef.get_schema(),
                    "title": _("Person references"),
                },
            },
        }

    def unserialize(self, data):
        """
        Convert the data held in a tuple created by the serialize method
        back into the data in a Person object.

        :param data: tuple containing the persistent data associated the
                     Person object
        :type data: tuple
        """
        (
            self.handle,  #  0
            self.gramps_id,  #  1
            self.__gender,  #  2
            primary_name,  #  3
            alternate_names,  #  4
            self.death_ref_index,  #  5
            self.birth_ref_index,  #  6
            event_ref_list,  #  7
            self.family_list,  #  8
            self.parent_family_list,  #  9
            media_list,  # 10
            address_list,  # 11
            attribute_list,  # 12
            urls,  # 13
            lds_ord_list,  # 14
            citation_list,  # 15
            note_list,  # 16
            self.change,  # 17
            tag_list,  # 18
            self.private,  # 19
            person_ref_list,  # 20
        ) = data

        self.primary_name = Name()
        self.primary_name.unserialize(primary_name)
        self.alternate_names = [Name().unserialize(name) for name in alternate_names]
        self.event_ref_list = [EventRef().unserialize(er) for er in event_ref_list]
        self.person_ref_list = [PersonRef().unserialize(pr) for pr in person_ref_list]
        MediaBase.unserialize(self, media_list)
        LdsOrdBase.unserialize(self, lds_ord_list)
        AddressBase.unserialize(self, address_list)
        AttributeBase.unserialize(self, attribute_list)
        UrlBase.unserialize(self, urls)
        CitationBase.unserialize(self, citation_list)
        NoteBase.unserialize(self, note_list)
        TagBase.unserialize(self, tag_list)
        return self

    def _has_handle_reference(self, classname, handle):
        """
        Return True if the object has reference to a given handle of given
        primary object type.

        :param classname: The name of the primary object class.
        :type classname: str
        :param handle: The handle to be checked.
        :type handle: str
        :returns: Returns whether the object has reference to this handle of
                  this object type.
        :rtype: bool
        """
        if classname == "Event":
            return any(ref.ref == handle for ref in self.event_ref_list)
        elif classname == "Person":
            return any(ref.ref == handle for ref in self.person_ref_list)
        elif classname == "Family":
            return any(
                ref == handle
                for ref in self.family_list
                + self.parent_family_list
                + [ordinance.famc for ordinance in self.lds_ord_list]
            )
        elif classname == "Place":
            return any(ordinance.place == handle for ordinance in self.lds_ord_list)
        return False

    def _remove_handle_references(self, classname, handle_list):
        if classname == "Event":
            # Keep a copy of the birth and death references
            birth_ref = self.get_birth_ref()
            death_ref = self.get_death_ref()

            new_list = [
                ref for ref in self.event_ref_list if ref.ref not in handle_list
            ]
            # If deleting removing the reference to the event
            # to which birth or death ref_index points, unset the index
            if (
                self.birth_ref_index != -1
                and self.event_ref_list[self.birth_ref_index].ref in handle_list
            ):
                self.set_birth_ref(None)
            if (
                self.death_ref_index != -1
                and self.event_ref_list[self.death_ref_index].ref in handle_list
            ):
                self.set_death_ref(None)
            self.event_ref_list = new_list

            # Reset the indexes after deleting the event from even_ref_list
            if self.birth_ref_index != -1:
                self.set_birth_ref(birth_ref)
            if self.death_ref_index != -1:
                self.set_death_ref(death_ref)
        elif classname == "Person":
            new_list = [
                ref for ref in self.person_ref_list if ref.ref not in handle_list
            ]
            self.person_ref_list = new_list
        elif classname == "Family":
            new_list = [
                handle for handle in self.family_list if handle not in handle_list
            ]
            self.family_list = new_list
            new_list = [
                handle
                for handle in self.parent_family_list
                if handle not in handle_list
            ]
            self.parent_family_list = new_list
            for ordinance in self.lds_ord_list:
                if ordinance.famc in handle_list:
                    ordinance.famc = None
        elif classname == "Place":
            for ordinance in self.lds_ord_list:
                if ordinance.place in handle_list:
                    ordinance.place = None

    def _replace_handle_reference(self, classname, old_handle, new_handle):
        if classname == "Event":
            refs_list = [ref.ref for ref in self.event_ref_list]
            new_ref = None
            if new_handle in refs_list:
                new_ref = self.event_ref_list[refs_list.index(new_handle)]
            n_replace = refs_list.count(old_handle)
            for ix_replace in range(n_replace):
                idx = refs_list.index(old_handle)
                self.event_ref_list[idx].ref = new_handle
                refs_list[idx] = new_handle
                if new_ref:
                    evt_ref = self.event_ref_list[idx]
                    equi = new_ref.is_equivalent(evt_ref)
                    if equi != DIFFERENT:
                        if equi == EQUAL:
                            new_ref.merge(evt_ref)
                        self.event_ref_list.pop(idx)
                        refs_list.pop(idx)
                        if idx < self.birth_ref_index:
                            self.birth_ref_index -= 1
                        elif idx == self.birth_ref_index:
                            self.birth_ref_index = -1
                            # birth_ref_index should be recalculated which
                            # needs database access!
                        if idx < self.death_ref_index:
                            self.death_ref_index -= 1
                        elif idx == self.death_ref_index:
                            self.death_ref_index = -1
                            # death_ref_index should be recalculated which
                            # needs database access!
        elif classname == "Person":
            refs_list = [ref.ref for ref in self.person_ref_list]
            new_ref = None
            if new_handle in refs_list:
                new_ref = self.person_ref_list[refs_list.index(new_handle)]
            n_replace = refs_list.count(old_handle)
            for ix_replace in range(n_replace):
                idx = refs_list.index(old_handle)
                self.person_ref_list[idx].ref = new_handle
                refs_list[idx] = new_handle
                if new_ref:
                    person_ref = self.person_ref_list[idx]
                    equi = new_ref.is_equivalent(person_ref)
                    if equi != DIFFERENT:
                        if equi == EQUAL:
                            new_ref.merge(person_ref)
                        self.person_ref_list.pop(idx)
                        refs_list.pop(idx)
        elif classname == "Family":
            while old_handle in self.family_list:
                ix = self.family_list.index(old_handle)
                self.family_list[ix] = new_handle
            while old_handle in self.parent_family_list:
                ix = self.parent_family_list.index(old_handle)
                self.parent_family_list[ix] = new_handle
            handle_list = [ordinance.famc for ordinance in self.lds_ord_list]
            while old_handle in handle_list:
                ix = handle_list.index(old_handle)
                self.lds_ord_list[ix].famc = new_handle
                handle_list[ix] = ""
        elif classname == "Place":
            handle_list = [ordinance.place for ordinance in self.lds_ord_list]
            while old_handle in handle_list:
                ix = handle_list.index(old_handle)
                self.lds_ord_list[ix].place = new_handle
                handle_list[ix] = ""

    def get_text_data_list(self):
        """
        Return the list of all textual attributes of the object.

        :returns: Returns the list of all textual attributes of the object.
        :rtype: list
        """
        return [self.gramps_id]

    def get_text_data_child_list(self):
        """
        Return the list of child objects that may carry textual data.

        :returns: Returns the list of child objects that may carry textual data.
        :rtype: list
        """
        check_list = self.lds_ord_list
        add_list = [_f for _f in check_list if _f]
        return (
            [self.primary_name]
            + self.media_list
            + self.alternate_names
            + self.address_list
            + self.attribute_list
            + self.urls
            + self.event_ref_list
            + add_list
            + self.person_ref_list
        )

    def get_citation_child_list(self):
        """
        Return the list of child secondary objects that may refer citations.

        :returns: Returns the list of child secondary child objects that may
                  refer citations.
        :rtype: list
        """
        return (
            [self.primary_name]
            + self.media_list
            + self.alternate_names
            + self.address_list
            + self.attribute_list
            + self.lds_ord_list
            + self.person_ref_list
            + self.event_ref_list
        )

    def get_note_child_list(self):
        """
        Return the list of child secondary objects that may refer notes.

        :returns: Returns the list of child secondary child objects that may
                  refer notes.
        :rtype: list
        """
        return (
            [self.primary_name]
            + self.media_list
            + self.alternate_names
            + self.address_list
            + self.attribute_list
            + self.lds_ord_list
            + self.person_ref_list
            + self.event_ref_list
        )

    def get_referenced_handles(self):
        """
        Return the list of (classname, handle) tuples for all directly
        referenced primary objects.

        :returns: List of (classname, handle) tuples for referenced objects.
        :rtype: list
        """
        return [
            ("Family", handle)
            for handle in (self.family_list + self.parent_family_list)
        ] + (
            self.get_referenced_note_handles()
            + self.get_referenced_citation_handles()
            + self.get_referenced_tag_handles()
        )

    def get_handle_referents(self):
        """
        Return the list of child objects which may, directly or through
        their children, reference primary objects.

        :returns: Returns the list of objects referencing primary objects.
        :rtype: list
        """
        return (
            [self.primary_name]
            + self.media_list
            + self.alternate_names
            + self.address_list
            + self.attribute_list
            + self.lds_ord_list
            + self.person_ref_list
            + self.event_ref_list
        )

    def merge(self, acquisition):
        """
        Merge the content of acquisition into this person.

        :param acquisition: The person to merge with the present person.
        :type acquisition: Person
        """
        acquisition_id = acquisition.get_gramps_id()
        if acquisition_id:
            attr = Attribute()
            attr.set_type(_("Merged Gramps ID"))
            attr.set_value(acquisition.get_gramps_id())
            self.add_attribute(attr)

        self._merge_privacy(acquisition)
        acquisition.alternate_names.insert(0, acquisition.get_primary_name())
        self._merge_alternate_names(acquisition)
        self._merge_event_ref_list(acquisition)
        self._merge_lds_ord_list(acquisition)
        self._merge_media_list(acquisition)
        self._merge_address_list(acquisition)
        self._merge_attribute_list(acquisition)
        self._merge_url_list(acquisition)
        self._merge_person_ref_list(acquisition)
        self._merge_note_list(acquisition)
        self._merge_citation_list(acquisition)
        self._merge_tag_list(acquisition)

        list(
            map(
                self.add_parent_family_handle,
                acquisition.get_parent_family_handle_list(),
            )
        )
        list(map(self.add_family_handle, acquisition.get_family_handle_list()))

    def set_primary_name(self, name):
        """
        Set the primary name of the Person to the specified :class:`~.name.Name`
        instance.

        :param name: :class:`~.name.Name` to be assigned to the person
        :type name: :class:`~.name.Name`
        """
        self.primary_name = name

    def get_primary_name(self):
        """
        Return the :class:`~.name.Name` instance marked as the Person's primary
        name.

        :returns: Returns the primary name
        :rtype: :class:`~.name.Name`
        """
        return self.primary_name

    def get_alternate_names(self):
        """
        Return the list of alternate :class:`~.name.Name` instances.

        :returns: List of :class:`~.name.Name` instances
        :rtype: list
        """
        return self.alternate_names

    def set_alternate_names(self, alt_name_list):
        """
        Change the list of alternate names to the passed list.

        :param alt_name_list: List of :class:`~.name.Name` instances
        :type alt_name_list: list
        """
        self.alternate_names = alt_name_list

    def _merge_alternate_names(self, acquisition):
        """
        Merge the list of alternate names from acquisition with our own.

        :param acquisition: the list of alternate names of this object will be
                            merged with the current alternate name list.
        :rtype acquisition: Person
        """
        name_list = self.alternate_names[:]
        primary_name = self.get_primary_name()
        if primary_name and not primary_name.is_empty():
            name_list.insert(0, primary_name)
        for addendum in acquisition.get_alternate_names():
            for name in name_list:
                equi = name.is_equivalent(addendum)
                if equi == IDENTICAL:
                    break
                elif equi == EQUAL:
                    name.merge(addendum)
                    break
            else:
                self.alternate_names.append(addendum)

    def add_alternate_name(self, name):
        """
        Add a :class:`~.name.Name` instance to the list of alternative names.

        :param name: :class:`~.name.Name` to add to the list
        :type name: :class:`~.name.Name`
        """
        self.alternate_names.append(name)

    def get_nick_name(self):
        for name in [self.get_primary_name()] + self.get_alternate_names():
            if name.get_nick_name():
                return name.get_nick_name()
        for attr in self.attribute_list:
            if int(attr.type) == AttributeType.NICKNAME:
                return attr.get_value()
        return ""

    def set_gender(self, gender):
        """
        Set the gender of the Person.

        :param gender: Assigns the Person's gender to one of the
                       following constants:

                       - Person.MALE
                       - Person.FEMALE
                       - Person.OTHER
                       - Person.UNKNOWN
        :type gender: int
        """
        if gender not in (Person.MALE, Person.FEMALE, Person.OTHER, Person.UNKNOWN):
            raise ValueError("Attempt to assign invalid gender")
        self.__gender = gender

    def get_gender(self):
        """
        Return the gender of the Person.

        :returns: Returns one of the following constants:

                  - Person.MALE
                  - Person.FEMALE
                  - Person.OTHER
                  - Person.UNKNOWN
        :rtype: int
        """
        return self.__gender

    gender = property(
        get_gender, set_gender, None, "Returns or sets the gender of the person"
    )

    def set_birth_ref(self, event_ref):
        """
        Assign the birth event to the Person object.

        This is accomplished by assigning the :class:`~.eventref.EventRef` of
        the birth event in the current database.

        :param event_ref: the :class:`~.eventref.EventRef` object associated
                          with the Person's birth.
        :type event_ref: EventRef
        """
        if event_ref and not isinstance(event_ref, EventRef):
            raise ValueError("Expecting EventRef instance")
        if event_ref is None:
            self.birth_ref_index = -1
            return

        # check whether we already have this ref in the list
        for self.birth_ref_index, ref in enumerate(self.event_ref_list):
            if event_ref.is_equal(ref):
                return  # Note: self.birth_ref_index already set

        self.event_ref_list.append(event_ref)
        self.birth_ref_index = len(self.event_ref_list) - 1

    def set_death_ref(self, event_ref):
        """
        Assign the death event to the Person object.

        This is accomplished by assigning the :class:`~.eventref.EventRef` of
        the death event in the current database.

        :param event_ref: the :class:`~.eventref.EventRef` object associated
                          with the Person's death.
        :type event_ref: EventRef
        """
        if event_ref and not isinstance(event_ref, EventRef):
            raise ValueError("Expecting EventRef instance")
        if event_ref is None:
            self.death_ref_index = -1
            return

        # check whether we already have this ref in the list
        for self.death_ref_index, ref in enumerate(self.event_ref_list):
            if event_ref.is_equal(ref):
                return  # Note: self.death_ref_index already set

        self.event_ref_list.append(event_ref)
        self.death_ref_index = len(self.event_ref_list) - 1

    def get_birth_ref(self):
        """
        Return the :class:`~.eventref.EventRef` for Person's birth event.

        This should correspond to an :class:`~.event.Event` in the database's
        :class:`~.event.Event` list.

        :returns: Returns the birth :class:`~.eventref.EventRef` or None if no
                  birth :class:`~.event.Event` has been assigned.
        :rtype: EventRef
        """

        if 0 <= self.birth_ref_index < len(self.event_ref_list):
            return self.event_ref_list[self.birth_ref_index]
        else:
            return None

    def get_death_ref(self):
        """
        Return the :class:`~.eventref.EventRef` for the Person's death event.

        This should correspond to an :class:`~.event.Event` in the database's
        :class:`~.event.Event` list.

        :returns: Returns the death :class:`~.eventref.EventRef` or None if no
                  death :class:`~.event.Event` has been assigned.
        :rtype: event_ref
        """

        if 0 <= self.death_ref_index < len(self.event_ref_list):
            return self.event_ref_list[self.death_ref_index]
        else:
            return None

    def add_event_ref(self, event_ref):
        """
        Add the :class:`~.eventref.EventRef` to the Person instance's
        :class:`~.eventref.EventRef` list.

        This is accomplished by assigning the :class:`~.eventref.EventRef` of a
        valid :class:`~.event.Event` in the current database.

        :param event_ref: the :class:`~.eventref.EventRef` to be added to the
                          Person's :class:`~.eventref.EventRef` list.
        :type event_ref: EventRef
        """
        if event_ref and not isinstance(event_ref, EventRef):
            raise ValueError("Expecting EventRef instance")

        # check whether we already have this ref in the list
        if not any(event_ref.is_equal(ref) for ref in self.event_ref_list):
            self.event_ref_list.append(event_ref)

    def get_event_ref_list(self):
        """
        Return the list of :class:`~.eventref.EventRef` objects associated with
        :class:`~.event.Event` instances.

        :returns: Returns the list of :class:`~.eventref.EventRef` objects
                  associated with the Person instance.
        :rtype: list
        """
        return self.event_ref_list

    def get_primary_event_ref_list(self):
        """
        Return the list of :class:`~.eventref.EventRef` objects associated with
        :class:`~.event.Event` instances that have been marked as primary
        events.

        :returns: Returns generator of :class:`~.eventref.EventRef` objects
                  associated with the Person instance.
        :rtype: generator
        """
        return (
            ref
            for ref in self.event_ref_list
            if ref.get_role() == EventRoleType.PRIMARY
        )

    def set_event_ref_list(self, event_ref_list):
        """
        Set the Person instance's :class:`~.eventref.EventRef` list to the
        passed list.

        :param event_ref_list: List of valid :class:`~.eventref.EventRef`
                               objects.
        :type event_ref_list: list
        """
        self.event_ref_list = event_ref_list

    def _merge_event_ref_list(self, acquisition):
        """
        Merge the list of event references from acquisition with our own.

        :param acquisition: the event references list of this object will be
                            merged with the current event references list.
        :rtype acquisition: Person
        """
        eventref_list = self.event_ref_list[:]
        for idx, addendum in enumerate(acquisition.get_event_ref_list()):
            for eventref in eventref_list:
                equi = eventref.is_equivalent(addendum)
                if equi == IDENTICAL:
                    break
                elif equi == EQUAL:
                    eventref.merge(addendum)
                    break
            else:
                self.event_ref_list.append(addendum)
                if self.birth_ref_index == -1 and idx == acquisition.birth_ref_index:
                    self.birth_ref_index = len(self.event_ref_list) - 1
                if self.death_ref_index == -1 and idx == acquisition.death_ref_index:
                    self.death_ref_index = len(self.event_ref_list) - 1

    def add_family_handle(self, family_handle):
        """
        Add the :class:`~.family.Family` handle to the Person instance's
        :class:`~.family.Family` list.

        This is accomplished by assigning the handle of a valid
        :class:`~.family.Family` in the current database.

        Adding a :class:`~.family.Family` handle to a Person does not
        automatically update the corresponding :class:`~.family.Family`. The
        developer is responsible to make sure that when a
        :class:`~.family.Family` is added to Person, that the Person is assigned
        to either the father or mother role in the :class:`~.family.Family`.

        :param family_handle: handle of the :class:`~.family.Family` to be added
                              to the Person's :class:`~.family.Family` list.
        :type family_handle: str
        """
        if family_handle not in self.family_list:
            self.family_list.append(family_handle)

    def set_preferred_family_handle(self, family_handle):
        """
        Set the family_handle specified to be the preferred
        :class:`~.family.Family`.

        The preferred :class:`~.family.Family` is determined by the first
        :class:`~.family.Family` in the :class:`~.family.Family` list, and is
        typically used to indicate the preferred :class:`~.family.Family` for
        navigation or reporting.

        The family_handle must already be in the list, or the function
        call has no effect.

        :param family_handle: Handle of the :class:`~.family.Family` to make the
                              preferred :class:`~.family.Family`.
        :type family_handle: str
        :returns: True if the call succeeded, False if the family_handle
                  was not already in the :class:`~.family.Family` list.
        :rtype: bool
        """
        if family_handle in self.family_list:
            self.family_list.remove(family_handle)
            self.family_list = [family_handle] + self.family_list
            return True
        else:
            return False

    def get_family_handle_list(self):
        """
        Return the list of :class:`~.family.Family` handles in which the person
        is a parent or spouse.

        :returns: Returns the list of handles corresponding to the
                  :class:`~.family.Family` records with which the person
                  is associated.
        :rtype: list
        """
        return self.family_list

    def set_family_handle_list(self, family_list):
        """
        Assign the passed list to the Person's list of families in which it is
        a parent or spouse.

        :param family_list: List of :class:`~.family.Family` handles to be
                            associated with the Person
        :type family_list: list
        """
        self.family_list = family_list

    def clear_family_handle_list(self):
        """
        Remove all :class:`~.family.Family` handles from the
        :class:`~.family.Family` list.
        """
        self.family_list = []

    def remove_family_handle(self, family_handle):
        """
        Remove the specified :class:`~.family.Family` handle from the list of
        marriages/partnerships.

        If the handle does not exist in the list, the operation has no effect.

        :param family_handle: :class:`~.family.Family` handle to remove from
                              the list
        :type family_handle: str

        :returns: True if the handle was removed, False if it was not
                  in the list.
        :rtype: bool
        """
        if family_handle in self.family_list:
            self.family_list.remove(family_handle)
            return True
        else:
            return False

    def get_parent_family_handle_list(self):
        """
        Return the list of :class:`~.family.Family` handles in which the person
        is a child.

        :returns: Returns the list of handles corresponding to the
                  :class:`~.family.Family` records with which the person is a
                  child.
        :rtype: list
        """
        return self.parent_family_list

    def set_parent_family_handle_list(self, family_list):
        """
        Return the list of :class:`~.family.Family` handles in which the person
        is a child.

        :returns: Returns the list of handles corresponding to the
                  :class:`~.family.Family` records with which the person is a
                  child.
        :rtype: list
        """
        self.parent_family_list = family_list

    def add_parent_family_handle(self, family_handle):
        """
        Add the :class:`~.family.Family` handle to the Person instance's list of
        families in which it is a child.

        This is accomplished by assigning the handle of a valid
        :class:`~.family.Family` in the current database.

        Adding a :class:`~.family.Family` handle to a Person does not
        automatically update the corresponding :class:`~.family.Family`. The
        developer is responsible to make sure that when a
        :class:`~.family.Family` is added to Person, that the Person is
        added to the :class:`~.family.Family` instance's child list.

        :param family_handle: handle of the :class:`~.family.Family` to be added
                              to the Person's :class:`~.family.Family` list.
        :type family_handle: str
        """
        if not isinstance(family_handle, str):
            raise ValueError("Expecting handle, obtained %s" % str(family_handle))
        if family_handle not in self.parent_family_list:
            self.parent_family_list.append(family_handle)

    def clear_parent_family_handle_list(self):
        """
        Remove all :class:`~.family.Family` handles from the parent
        :class:`~.family.Family` list.
        """
        self.parent_family_list = []

    def remove_parent_family_handle(self, family_handle):
        """
        Remove the specified :class:`~.family.Family` handle from the list of
        parent families (families in which the parent is a child).

        If the handle does not exist in the list, the operation has no effect.

        :param family_handle: :class:`~.family.Family` handle to remove from the
                              list
        :type family_handle: str

        :returns: Returns a tuple of three strings, consisting of the
                  removed handle, relationship to mother, and relationship
                  to father. None is returned if the handle is not in the
                  list.
        :rtype: tuple
        """
        if family_handle in self.parent_family_list:
            self.parent_family_list.remove(family_handle)
            return True
        else:
            return False

    def set_main_parent_family_handle(self, family_handle):
        """
        Set the main :class:`~.family.Family` in which the Person is a child.

        The main :class:`~.family.Family` is the :class:`~.family.Family`
        typically used for reports and navigation. This is accomplished by
        moving the :class:`~.family.Family` to the beginning of the list. The
        family_handle must be in the list for this to have any effect.

        :param family_handle: handle of the :class:`~.family.Family` to be
                              marked as the main :class:`~.family.Family`
        :type family_handle: str
        :returns: Returns True if the assignment has successful
        :rtype: bool
        """
        if family_handle in self.parent_family_list:
            self.parent_family_list.remove(family_handle)
            self.parent_family_list = [family_handle] + self.parent_family_list
            return True
        else:
            return False

    def get_main_parents_family_handle(self):
        """
        Return the handle of the :class:`~.family.Family` considered to be the
        main :class:`~.family.Family` in which the Person is a child.

        :returns: Returns the family_handle if a family_handle exists,
                  If no :class:`~.family.Family` is assigned, None is returned
        :rtype: str
        """
        if self.parent_family_list:
            return self.parent_family_list[0]
        else:
            return None

    def add_person_ref(self, person_ref):
        """
        Add the :class:`~.personref.PersonRef` to the Person instance's
        :class:`~.personref.PersonRef` list.

        :param person_ref: the :class:`~.personref.PersonRef` to be added to the
                           Person's :class:`~.personref.PersonRef` list.
        :type person_ref: PersonRef
        """
        if person_ref and not isinstance(person_ref, PersonRef):
            raise ValueError("Expecting PersonRef instance")
        self.person_ref_list.append(person_ref)

    def get_person_ref_list(self):
        """
        Return the list of :class:`~.personref.PersonRef` objects.

        :returns: Returns the list of :class:`~.personref.PersonRef` objects.
        :rtype: list
        """
        return self.person_ref_list

    def set_person_ref_list(self, person_ref_list):
        """
        Set the Person instance's :class:`~.personref.PersonRef` list to the
        passed list.

        :param person_ref_list: List of valid :class:`~.personref.PersonRef`
                                objects
        :type person_ref_list: list
        """
        self.person_ref_list = person_ref_list

    def _merge_person_ref_list(self, acquisition):
        """
        Merge the list of person references from acquisition with our own.

        :param acquisition: the list of person references of this person will be
                            merged with the current person references list.
        :rtype acquisition: Person
        """
        personref_list = self.person_ref_list[:]
        for addendum in acquisition.get_person_ref_list():
            for personref in personref_list:
                equi = personref.is_equivalent(addendum)
                if equi == IDENTICAL:
                    break
                elif equi == EQUAL:
                    personref.merge(addendum)
                    break
            else:
                self.person_ref_list.append(addendum)
